package org.etsi.osl.tmf.pm628.reposervices;


import jakarta.validation.Valid;
import org.etsi.osl.tmf.pm628.model.DataAccessEndpoint;
import org.etsi.osl.tmf.pm628.repo.DataAccessEndpointRepository;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import java.util.List;
import java.util.Optional;

@Service
@Transactional
public class DataAccessEndpointService {

    private static final Logger log = LoggerFactory.getLogger(DataAccessEndpointService.class);

    private final DataAccessEndpointRepository dataAccessEndpointRepository;

    @Autowired
    public DataAccessEndpointService(DataAccessEndpointRepository dataAccessEndpointRepository) {
        this.dataAccessEndpointRepository = dataAccessEndpointRepository;
    }

    public List<DataAccessEndpoint> findAllDataAccessEndpoints(){
        log.debug("DataAccessEndpoint LIST");
        return (List<DataAccessEndpoint>) dataAccessEndpointRepository.findAll();
    }

    public DataAccessEndpoint findDataAccessEndpointByUuid(String uuid){
        log.debug("DataAccessEndpoint FIND BY UUID");
        Optional<DataAccessEndpoint> dataAccessEndpoint = dataAccessEndpointRepository.findByUuid(uuid);
        return dataAccessEndpoint.orElse(null);
    }

    public DataAccessEndpoint createDataAccessEndpoint(DataAccessEndpoint dataAccessEndpoint){
        log.info("DataAccessEndpoint CREATE: {}",dataAccessEndpoint);
        return dataAccessEndpointRepository.save(dataAccessEndpoint);
    }

    public DataAccessEndpoint updateDataAccessEndpoint(String uuid, @Valid DataAccessEndpoint dataAccessEndpointUpdate){
        log.info("DataAccessEndpoint UPDATE with UUID: {}", uuid);
        DataAccessEndpoint dataAccessEndpoint = dataAccessEndpointRepository.findByUuid(uuid).
                orElseThrow(() -> new IllegalArgumentException("No Data Access Endpoint with UUID: " + uuid));
        dataAccessEndpoint = updateDataAccessEndpointFields(dataAccessEndpoint, dataAccessEndpointUpdate);
        dataAccessEndpointRepository.save(dataAccessEndpoint);
        return dataAccessEndpoint;
    }

    public void  deleteDataAccessEndpoint(String uuid){
        log.info("DataAccessEndpoint DELETE with UUID:{}", uuid);
        DataAccessEndpoint dataAccessEndpoint = dataAccessEndpointRepository.findByUuid(uuid)
                .orElseThrow(() -> new IllegalArgumentException("No Data Access Endpoint with UUID: " + uuid));
        dataAccessEndpointRepository.delete(dataAccessEndpoint);
    }

    /**
     * Updates the fields of the current DataAccessEndpoint object with the values from the update DataAccessEndpoint object.
     * Only the non-null fields of the update DataAccessEndpoint object are used.
     * This method is typically used for updating an existing DataAccessEndpoint with new values.
     *
     * @param daeCurrent The current DataAccessEndpoint object that is to be updated.
     * @param daeUpdate The DataAccessEndpoint object that contains the new values.
     * @return The updated DataAccessEndpoint object.
     *
     * @throws IllegalArgumentException if the provided update DataAccessEndpoint object is not valid.
     */
    public DataAccessEndpoint updateDataAccessEndpointFields(DataAccessEndpoint daeCurrent, @Valid DataAccessEndpoint daeUpdate){
        if (daeUpdate.getApiType() != null) {
            daeCurrent.setApiType(
                    daeUpdate.getApiType()
            );
        }
        if (daeUpdate.getUri() != null) {
            daeCurrent.setUri(
                    daeUpdate.getUri()
            );
        }
        if (daeUpdate.getUriQueryFilter() != null) {
            daeCurrent.setUriQueryFilter(
                    daeUpdate.getUriQueryFilter()
            );
        }
        if (daeUpdate.getDescription() != null) {
            daeCurrent.setDescription(
                    daeUpdate.getDescription()
            );
        }
        if (daeUpdate.getCategory() != null) {
            daeCurrent.setCategory(
                    daeUpdate.getCategory()
            );
        }
        if (daeUpdate.getActivationFeature() != null) {
            daeCurrent.setActivationFeature(
                    daeUpdate.getActivationFeature()
            );
        }
        if (daeUpdate.getAdministrativeState() != null) {
            daeCurrent.setAdministrativeState(
                    daeUpdate.getAdministrativeState()
            );
        }
        if (daeUpdate.getAtBaseType() != null) {
            daeCurrent.setAtBaseType(
                    daeUpdate.getAtBaseType()
            );
        }
        if (daeUpdate.getAtSchemaLocation() != null) {
            daeCurrent.setAtSchemaLocation(
                    daeUpdate.getAtSchemaLocation()
            );
        }
        if (daeUpdate.getAttachment() != null) {
            daeCurrent.setAttachment(
                    daeUpdate.getAttachment()
            );
        }
        if (daeUpdate.getAtType() != null) {
            daeCurrent.setApiType(
                    daeUpdate.getAtType()
            );
        }
        if (daeUpdate.getEndOperatingDate() != null) {
            daeCurrent.setEndOperatingDate(
                    daeUpdate.getEndOperatingDate()
            );
        }
        if (daeUpdate.getExternalIdentifier() != null) {
            daeCurrent.setExternalIdentifier(
                    daeUpdate.getExternalIdentifier()
            );
        }
        if (daeUpdate.getHref() != null) {
            daeCurrent.setHref(
                    daeUpdate.getHref()
            );
        }
        if (daeUpdate.getId() != null) {
            daeCurrent.setId(
                    daeUpdate.getId()
            );
        }
        if (daeUpdate.getIntent() != null) {
            daeCurrent.setIntent(
                    daeUpdate.getIntent()
            );
        }
        if (daeUpdate.getName() != null) {
            daeCurrent.setName(
                    daeUpdate.getName()
            );
        }
        if (daeUpdate.getNote() != null) {
            daeCurrent.setNote(
                    daeUpdate.getNote()
            );
        }
        if (daeUpdate.getOperationalState() != null) {
            daeCurrent.setOperationalState(
                    daeUpdate.getOperationalState()
            );
        }
        if (daeUpdate.getPlace() != null) {
            daeCurrent.setPlace(
                    daeUpdate.getPlace()
            );
        }
        if (daeUpdate.getRelatedParty() != null) {
            daeCurrent.setRelatedParty(
                    daeUpdate.getRelatedParty()
            );
        }
        if (daeUpdate.getResourceCharacteristic() != null) {
            daeCurrent.setResourceCharacteristic(
                    daeUpdate.getResourceCharacteristic()
            );
        }
        if (daeUpdate.getResourceOrderItem() != null) {
            daeCurrent.setResourceOrderItem(
                    daeUpdate.getResourceOrderItem()
            );
        }
        if (daeUpdate.getResourceRelationship() != null) {
            daeCurrent.setResourceRelationship(
                    daeUpdate.getResourceRelationship()
            );
        }
        if (daeUpdate.getResourceStatus() != null) {
            daeCurrent.setResourceStatus(
                    daeUpdate.getResourceStatus()
            );
        }
        if (daeUpdate.getResourceSpecification() != null) {
            daeCurrent.setResourceSpecification(
                    daeUpdate.getResourceSpecification()
            );
        }
        if (daeUpdate.getResourceVersion() != null) {
            daeCurrent.setResourceVersion(
                    daeUpdate.getResourceVersion()
            );
        }
        if (daeUpdate.getStartOperatingDate() != null) {
            daeCurrent.setStartOperatingDate(
                    daeUpdate.getStartOperatingDate()
            );
        }
        if (daeUpdate.getUsageState() != null) {
            daeCurrent.setUsageState(
                    daeUpdate.getUsageState()
            );
        }
        if (daeUpdate.getValidFor() != null) {
            daeCurrent.setValidFor(
                    daeUpdate.getValidFor()
            );
        }


        return daeCurrent;
    }

}
